@@ -1,4 +1,4 @@
-/* $Id: keyboard-drawing.c,v 1.39 2005/04/10 01:18:52 svu Exp $ */
+/* $Id: keyboard-drawing.c,v 1.44 2005/09/27 00:38:44 svu Exp $ */
 /*
  * keyboard-drawing.c: implementation of a gtk+ widget that is a drawing of
  * the keyboard of the default display
@@ -24,10 +24,10 @@
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
 #include <gdk/gdkkeysyms.h>
-#include <pango/pangoxft.h>
 #include <X11/XKBlib.h>
 #include <X11/extensions/XKBgeom.h>
 #include <stdlib.h>
+#include <memory.h>
 #include <math.h>
 #include "keyboard-drawing.h"
 #include "keyboard-marshal.h"
@@ -43,7 +43,7 @@ enum
 static guint keyboard_drawing_signals[NUM_SIGNALS] = { 0 };
 
 static void keyboard_drawing_set_mods (
-  KeyboardDrawing * kbdrawing,
+  KeyboardDrawing * drawing,
   guint mods);
 
 static gint
@@ -167,16 +167,19 @@ draw_rectangle (
   else
     {
       XkbPointRec points[4];
+      gint x, y;
 
       points[0].x = xkb_x;
       points[0].y = xkb_y;
-      rotate_coordinate (xkb_x, xkb_y, xkb_x + xkb_width, xkb_y, angle,
-                         (gint *) & points[1].x, (gint *) & points[1].y);
+      rotate_coordinate (xkb_x, xkb_y, xkb_x + xkb_width, xkb_y, 
+                         angle, &x, &y);
+      points[1].x = x; points[1].y = y;
       rotate_coordinate (xkb_x, xkb_y, xkb_x + xkb_width, xkb_y + xkb_height,
-                         angle, (gint *) & points[2].x,
-                         (gint *) & points[2].y);
-      rotate_coordinate (xkb_x, xkb_y, xkb_x, xkb_y + xkb_height, angle,
-                         (gint *) & points[3].x, (gint *) & points[3].y);
+                         angle, &x, &y);
+      points[2].x = x; points[2].y = y;
+      rotate_coordinate (xkb_x, xkb_y, xkb_x, xkb_y + xkb_height, 
+                         angle, &x, &y);
+      points[3].x = x; points[3].y = y;
 
       /* the points we've calculated are relative to 0,0 */
       draw_polygon (drawing, fill_color, 0, 0, points, 4);
@@ -192,12 +195,21 @@ draw_outline (
   gint origin_x,
   gint origin_y)
 {
+#ifdef KBDRAW_DEBUG
+  printf("num_points in %p: %d\n", outline, outline->num_points);
+#endif
+
   if (outline->num_points == 1)
     {
       if (color)
         draw_rectangle (drawing, color, angle, origin_x, origin_y,
                         outline->points[0].x, outline->points[0].y);
 
+#ifdef KBDRAW_DEBUG
+      printf("points:%p\n", outline->points);
+      printf("pointsxy:%d %d\n", outline->points[0].x, outline->points[0].y);
+#endif
+
       draw_rectangle (drawing, NULL, angle, origin_x, origin_y,
                       outline->points[0].x, outline->points[0].y);
     }
@@ -312,6 +324,8 @@ find_keycode (
 {
   guint i;
 
+  if (!drawing->xkb) return (gint)(-1);
+
   for (i = drawing->xkb->min_key_code; i <= drawing->xkb->max_key_code; i++)
     {
       if (drawing->xkb->names->keys[i].name[0] == key_name[0] &&
@@ -525,12 +539,37 @@ is_key_flipped (
   KeyboardDrawing * drawing,
   guint keycode)
 {
-  gint level, group;
+  gint level, group, gw, lw;
+  /*uint mods_rtrn;*/
+  /*KeySym keysym;*/
 
-  for (group = 0; group < XkbKeyGroupsWidth (drawing->xkb, keycode); group++)
-    for (level = 0; level < XkbKeyGroupWidth (drawing->xkb, keycode, group);
-         level++)
-      switch (XkbKeySymEntry (drawing->xkb, keycode, level, group))
+#ifdef KBDRAW_DEBUG
+  printf ("flipped of %04d: ", keycode);
+#endif
+
+  if (!drawing->xkb) return FALSE;
+
+  /* temporary solution */
+  return FALSE;
+
+  gw = XkbKeyGroupsWidth (drawing->xkb, keycode);
+#ifdef KBDRAW_DEBUG
+  printf("gw: %d ", gw);
+#endif
+
+  for (group = 0; group < gw; group++)
+  {
+    lw = XkbKeyGroupWidth (drawing->xkb, keycode, group);
+#ifdef KBDRAW_DEBUG
+    printf("lw: %d ", lw);
+#endif
+    for (level = 0; level < lw; level++)
+    {
+      KeySym ks = XkbKeySymEntry (drawing->xkb, keycode, level, group);
+#ifdef KBDRAW_DEBUG
+      printf ("ks[g:%d,l:%d]: %04d(%c) ", group, level, (int)ks, (char)ks);
+#endif
+      switch (ks)
         {
         case GDK_Escape:
         case GDK_F1:
@@ -610,27 +649,20 @@ is_key_flipped (
         case GDK_KP_Right:
         case GDK_KP_Up:
         case GDK_KP_Down:
+#ifdef KBDRAW_DEBUG
+  printf ("true\n");
+#endif
           return TRUE;
         }
+    }
+  }
 
+#ifdef KBDRAW_DEBUG
+  printf ("false\n");
+#endif
   return FALSE;
 }
 
-static void
-substitute_func_diagonal_baseline (
-  FcPattern * pattern,
-  gpointer data)
-{
-  gint angle = GPOINTER_TO_INT (data);
-  FcMatrix matrix;
-  FcBool ret;
-
-  FcMatrixInit (&matrix);
-  FcMatrixRotate (&matrix, cos (-M_PI * angle / 1800.0),
-                  sin (-M_PI * angle / 1800.0));
-  ret = FcPatternAddMatrix (pattern, FC_MATRIX, &matrix);
-}
-
 static void
 draw_layout (
   KeyboardDrawing * drawing,
@@ -649,24 +681,9 @@ draw_layout (
 
   if (angle != drawing->angle)
     {
-      if (angle == 0)
-        {
-          pango_xft_set_default_substitute (drawing->display,
-                                            drawing->screen_num, NULL, NULL,
-                                            NULL);
-          pango_xft_substitute_changed (drawing->display,
-                                        drawing->screen_num);
-        }
-      else
-        {
-          pango_xft_set_default_substitute (drawing->display,
-                                            drawing->screen_num,
-                                            substitute_func_diagonal_baseline,
-                                            GINT_TO_POINTER (angle), NULL);
-          pango_xft_substitute_changed (drawing->display,
-                                        drawing->screen_num);
-        }
-
+      PangoMatrix matrix = PANGO_MATRIX_INIT;
+      pango_matrix_rotate (&matrix, -angle / 10.0);
+      pango_context_set_matrix (gtk_widget_get_pango_context (GTK_WIDGET (drawing)), &matrix);
       pango_layout_context_changed (drawing->layout);
       drawing->angle = angle;
     }
@@ -729,6 +746,10 @@ draw_key_label_helper (
 
   if (keysym == 0)
     return;
+#ifdef KBDRAW_DEBUG
+  printf ("keysym: %04X(%c) at glp: %d, flipped: %d\n",
+          (unsigned)keysym, (char)keysym, (int)glp, flipped);
+#endif
 
   switch( glp )
   {
@@ -791,6 +812,8 @@ draw_key_label (
   gint padding;
   gint g, l, glp;
 
+  if (!drawing->xkb) return;
+
   padding = 23 * drawing->scale_numerator / drawing->scale_denominator; /* 2.3mm */
 
   x = xkb_to_pixmap_coord (drawing, xkb_origin_x);
@@ -806,6 +829,7 @@ draw_key_label (
         continue;
       g = drawing->groupLevels[glp]->group;
       l = drawing->groupLevels[glp]->level;
+
       if (g < 0 || g >= XkbKeyNumGroups (drawing->xkb, keycode))
         continue;
       if (l < 0 || l >= XkbKeyGroupWidth (drawing->xkb, keycode, g))
@@ -847,15 +871,28 @@ draw_key (
   GdkColor *color;
   gint i;
 
-  shape = &drawing->xkb->geom->shapes[key->xkbkey->shape_ndx];
+  if (!drawing->xkb) return;
+
+#ifdef KBDRAW_DEBUG
+  printf ("shape: %p (%p + %d)\n",
+          drawing->xkb->geom->shapes + key->xkbkey->shape_ndx,
+          drawing->xkb->geom->shapes,
+          key->xkbkey->shape_ndx);
+#endif
+
+  shape = drawing->xkb->geom->shapes + key->xkbkey->shape_ndx;
 
   if (key->pressed)
     color = &(GTK_WIDGET (drawing)->style->base[GTK_STATE_SELECTED]);
   else
-    color = &drawing->colors[key->xkbkey->color_ndx];
+    color = drawing->colors + key->xkbkey->color_ndx;
+
+#ifdef KBDRAW_DEBUG
+  printf ("outlines: %p(%d)\n", shape->outlines, shape->num_outlines);
+#endif
 
   for (i = 0; i < 1 /* shape->num_outlines */ ; i++)
-    draw_outline (drawing, &shape->outlines[i], color, key->angle,
+    draw_outline (drawing, shape->outlines + i, color, key->angle,
                   key->origin_x, key->origin_y);
 
   draw_key_label (drawing, key->keycode, key->angle, key->origin_x,
@@ -873,14 +910,16 @@ invalidate_region (
   GdkPoint points[4];
   gint x_min, x_max, y_min, y_max;
   gint x, y, width, height;
+  gint xx, yy;
 
-  rotate_coordinate (0, 0, 0, 0, angle, &points[0].x, &points[0].y);
-  rotate_coordinate (0, 0, shape->bounds.x2, 0, angle, &points[1].x,
-                     &points[1].y);
-  rotate_coordinate (0, 0, shape->bounds.x2, shape->bounds.y2, angle,
-                     &points[2].x, &points[2].y);
-  rotate_coordinate (0, 0, 0, shape->bounds.y2, angle, &points[3].x,
-                     &points[3].y);
+  rotate_coordinate (0, 0, 0, 0, angle, &xx, &yy);
+  points[0].x = xx; points[0].y = yy;
+  rotate_coordinate (0, 0, shape->bounds.x2, 0, angle, &xx, &yy);
+  points[1].x = xx; points[1].y = yy;
+  rotate_coordinate (0, 0, shape->bounds.x2, shape->bounds.y2, angle, &xx, &yy);
+  points[2].x = xx; points[2].y = yy;
+  rotate_coordinate (0, 0, 0, shape->bounds.y2, angle, &xx, &yy);
+  points[3].x = xx; points[3].y = yy;
 
   x_min =
     MIN (MIN (points[0].x, points[1].x), MIN (points[2].x, points[3].x));
@@ -904,6 +943,8 @@ invalidate_indicator_doodad_region (
   KeyboardDrawing * drawing,
   KeyboardDrawingDoodad * doodad)
 {
+  if (!drawing->xkb) return;
+
   invalidate_region (drawing,
                      doodad->angle,
                      doodad->origin_x + doodad->doodad->indicator.left,
@@ -917,6 +958,8 @@ invalidate_key_region (
   KeyboardDrawing * drawing,
   KeyboardDrawingKey * key)
 {
+  if (!drawing->xkb) return;
+
   invalidate_region (drawing,
                      key->angle,
                      key->origin_x,
@@ -930,9 +973,11 @@ draw_text_doodad (
   KeyboardDrawingDoodad * doodad,
   XkbTextDoodadRec * text_doodad)
 {
-  gint x =
-    xkb_to_pixmap_coord (drawing, doodad->origin_x + text_doodad->left);
-  gint y = xkb_to_pixmap_coord (drawing, doodad->origin_y + text_doodad->top);
+  gint x, y;
+  if (!drawing->xkb) return;
+
+  x = xkb_to_pixmap_coord (drawing, doodad->origin_x + text_doodad->left);
+  y = xkb_to_pixmap_coord (drawing, doodad->origin_y + text_doodad->top);
 
   pango_layout_set_text (drawing->layout, text_doodad->text, -1);
   draw_layout (drawing, doodad->angle, x, y, drawing->layout);
@@ -948,14 +993,16 @@ draw_indicator_doodad (
   XkbShapeRec *shape;
   gint i;
 
-  shape = &drawing->xkb->geom->shapes[indicator_doodad->shape_ndx];
-  if (doodad->on)
-    color = &drawing->colors[indicator_doodad->on_color_ndx];
-  else
-    color = &drawing->colors[indicator_doodad->off_color_ndx];
+  if (!drawing->xkb) return;
+
+  shape = drawing->xkb->geom->shapes + indicator_doodad->shape_ndx;
+
+  color = drawing->colors + (doodad->on ?
+    indicator_doodad->on_color_ndx :
+    indicator_doodad->off_color_ndx);
 
   for (i = 0; i < 1; i++)
-    draw_outline (drawing, &shape->outlines[i], color, doodad->angle,
+    draw_outline (drawing, shape->outlines + i, color, doodad->angle,
                   doodad->origin_x + indicator_doodad->left,
                   doodad->origin_y + indicator_doodad->top);
 }
@@ -970,11 +1017,13 @@ draw_shape_doodad (
   GdkColor *color;
   gint i;
 
-  shape = &drawing->xkb->geom->shapes[shape_doodad->shape_ndx];
-  color = &drawing->colors[shape_doodad->color_ndx];
+  if (!drawing->xkb) return;
+
+  shape = drawing->xkb->geom->shapes + shape_doodad->shape_ndx;
+  color = drawing->colors + shape_doodad->color_ndx;
 
   for (i = 0; i < shape->num_outlines; i++)
-    draw_outline (drawing, &shape->outlines[i], color, doodad->angle,
+    draw_outline (drawing, shape->outlines + i, color, doodad->angle,
                   doodad->origin_x + shape_doodad->left,
                   doodad->origin_y + shape_doodad->top);
 }
@@ -1012,6 +1061,8 @@ draw_keyboard_item (
   KeyboardDrawingItem * item,
   KeyboardDrawing * drawing)
 {
+  if (!drawing->xkb) return;
+
   switch (item->type)
     {
     case KEYBOARD_DRAWING_ITEM_TYPE_KEY:
@@ -1031,6 +1082,8 @@ draw_keyboard (
   GtkStateType state = GTK_WIDGET_STATE (GTK_WIDGET (drawing));
   gint pixw, pixh;
 
+  if (!drawing->xkb) return;
+
   pixw = GTK_WIDGET (drawing)->allocation.width;
   pixh = GTK_WIDGET (drawing)->allocation.height;
 
@@ -1046,8 +1099,6 @@ draw_keyboard (
     return;
 
 #ifdef KBDRAW_DEBUG
-  printf ("groups: %d, %d\n", drawing->leftgroup, drawing->rightgroup);
-  printf ("levels: %d, %d\n", drawing->bottomlevel, drawing->toplevel);
   printf ("mods: %d\n", drawing->mods);
 #endif
 
@@ -1055,6 +1106,22 @@ draw_keyboard (
                   drawing);
 }
 
+static void 
+alloc_pango_layout (
+  KeyboardDrawing * drawing)
+{
+  PangoContext *context = gtk_widget_get_pango_context (GTK_WIDGET (drawing));
+  drawing->layout = pango_layout_new (context);
+}
+
+static void
+free_pango_layout (
+  KeyboardDrawing * drawing)
+{
+  g_object_unref (G_OBJECT (drawing->layout));
+  drawing->layout = NULL;
+}
+
 static gboolean
 expose_event (
   GtkWidget * widget,
@@ -1063,6 +1130,8 @@ expose_event (
 {
   GtkStateType state = GTK_WIDGET_STATE (GTK_WIDGET (drawing));
 
+  if (!drawing->xkb) return FALSE;
+
   if (drawing->pixmap == NULL)
     draw_keyboard (drawing);
 
@@ -1089,6 +1158,8 @@ size_allocate (
   GtkAllocation * allocation,
   KeyboardDrawing * drawing)
 {
+  if (!drawing->xkb) return;
+
   if (drawing->pixmap)
     {
       g_object_unref (drawing->pixmap);
@@ -1128,7 +1199,10 @@ key_event (
   GdkEventKey * event,
   KeyboardDrawing * drawing)
 {
-  KeyboardDrawingKey *key = &drawing->keys[event->hardware_keycode];
+  KeyboardDrawingKey *key;
+  if (!drawing->xkb) return FALSE;
+
+  key = drawing->keys + event->hardware_keycode;
 
   if (event->hardware_keycode > drawing->xkb->max_key_code ||
       event->hardware_keycode < drawing->xkb->min_key_code ||
@@ -1147,6 +1221,7 @@ key_event (
   key->pressed = (event->type == GDK_KEY_PRESS);
 
   draw_key (drawing, key);
+
   invalidate_key_region (drawing, key);
 
   return FALSE;
@@ -1158,6 +1233,8 @@ button_press_event (
   GdkEventButton * event,
   KeyboardDrawing * drawing)
 {
+  if (!drawing->xkb) return FALSE;
+
   gtk_widget_grab_focus (widget);
   return FALSE;
 }
@@ -1168,12 +1245,14 @@ unpress_keys (
 {
   gint i;
 
+  if (!drawing->xkb) return FALSE;
+
   for (i = drawing->xkb->min_key_code; i <= drawing->xkb->max_key_code; i++)
     if (drawing->keys[i].pressed)
       {
         drawing->keys[i].pressed = FALSE;
-        draw_key (drawing, &drawing->keys[i]);
-        invalidate_key_region (drawing, &drawing->keys[i]);
+        draw_key (drawing, drawing->keys + i);
+        invalidate_key_region (drawing, drawing->keys + i);
       }
 
   return FALSE;
@@ -1216,6 +1295,8 @@ init_indicator_doodad (
   XkbDoodadRec * xkbdoodad,
   KeyboardDrawingDoodad * doodad)
 {
+  if (!drawing->xkb) return;
+
   if (xkbdoodad->any.type == XkbIndicatorDoodad)
     {
       gint index;
@@ -1263,9 +1344,11 @@ init_keys_and_doodads (
   gint i, j, k;
   gint x, y;
 
+  if (!drawing->xkb) return;
+
   for (i = 0; i < drawing->xkb->geom->num_doodads; i++)
     {
-      XkbDoodadRec *xkbdoodad = &drawing->xkb->geom->doodads[i];
+      XkbDoodadRec *xkbdoodad = drawing->xkb->geom->doodads + i;
       KeyboardDrawingDoodad *doodad = g_new (KeyboardDrawingDoodad, 1);
 
       doodad->type = KEYBOARD_DRAWING_ITEM_TYPE_DOODAD;
@@ -1283,7 +1366,10 @@ init_keys_and_doodads (
 
   for (i = 0; i < drawing->xkb->geom->num_sections; i++)
     {
-      XkbSectionRec *section = &drawing->xkb->geom->sections[i];
+#ifdef KBDRAW_DEBUG
+      printf ("initing section %d\n", i);
+#endif
+      XkbSectionRec *section = drawing->xkb->geom->sections + i;
       guint priority;
 
       x = section->left;
@@ -1292,19 +1378,28 @@ init_keys_and_doodads (
 
       for (j = 0; j < section->num_rows; j++)
         {
-          XkbRowRec *row = &section->rows[j];
+          XkbRowRec *row = section->rows + j;
 
+#ifdef KBDRAW_DEBUG
+          printf ("  initing row %d\n", j);
+#endif
           x = section->left + row->left;
           y = section->top + row->top;
 
           for (k = 0; k < row->num_keys; k++)
             {
-              XkbKeyRec *xkbkey = &row->keys[k];
+              XkbKeyRec *xkbkey = row->keys + k;
               KeyboardDrawingKey *key;
               XkbShapeRec *shape =
-                &drawing->xkb->geom->shapes[xkbkey->shape_ndx];
+                drawing->xkb->geom->shapes + xkbkey->shape_ndx;
               guint keycode = find_keycode (drawing, xkbkey->name.name);
 
+#ifdef KBDRAW_DEBUG
+              printf ("    initing key %d, shape: %p(%p + %d), code: %d\n", k,
+                shape,
+                drawing->xkb->geom->shapes,
+                xkbkey->shape_ndx, keycode);
+#endif
               if (row->vertical)
                 y += xkbkey->gap;
               else
@@ -1312,7 +1407,7 @@ init_keys_and_doodads (
 
               if (keycode >= drawing->xkb->min_key_code &&
                   keycode <= drawing->xkb->max_key_code)
-                key = &drawing->keys[keycode];
+                key = drawing->keys + keycode;
               else
                 {
                   g_warning ("key %4.4s: keycode = %u; not in range %d..%d\n",
@@ -1346,7 +1441,7 @@ init_keys_and_doodads (
 
       for (j = 0; j < section->num_doodads; j++)
         {
-          XkbDoodadRec *xkbdoodad = &section->doodads[j];
+          XkbDoodadRec *xkbdoodad = section->doodads + j;
           KeyboardDrawingDoodad *doodad = g_new (KeyboardDrawingDoodad, 1);
 
           doodad->type = KEYBOARD_DRAWING_ITEM_TYPE_DOODAD;
@@ -1374,13 +1469,15 @@ init_colors (
   gboolean result;
   gint i;
 
+  if (!drawing->xkb) return;
+
   drawing->colors = g_new (GdkColor, drawing->xkb->geom->num_colors);
 
   for (i = 0; i < drawing->xkb->geom->num_colors; i++)
     {
       result =
         parse_xkb_color_spec (drawing->xkb->geom->colors[i].spec,
-                              &drawing->colors[i]);
+                              drawing->colors + i);
 
       if (!result)
         g_warning ("init_colors: unable to parse color %s\n",
@@ -1389,11 +1486,13 @@ init_colors (
 }
 
 static void
-free_keys_and_doodads_and_colors (
+free_cdik ( /*colors doodads indicators keys*/
   KeyboardDrawing * drawing)
 {
   GList *itemp;
 
+  if (!drawing->xkb) return;
+
   for (itemp = drawing->keyboard_items; itemp; itemp = itemp->next)
     {
       KeyboardDrawingItem *item = itemp->data;
@@ -1422,6 +1521,17 @@ free_keys_and_doodads_and_colors (
   g_free (drawing->colors);
 }
 
+static void
+alloc_cdik (
+  KeyboardDrawing * drawing)
+{
+  drawing->physical_indicators_size =
+    drawing->xkb->indicators->phys_indicators + 1;
+  drawing->physical_indicators =
+    g_new0 (KeyboardDrawingDoodad *, drawing->physical_indicators_size);
+  drawing->keys = g_new0 (KeyboardDrawingKey, drawing->xkb->max_key_code + 1);
+}
+
 static GdkFilterReturn
 xkb_state_notify_event_filter (
   GdkXEvent * gdkxev,
@@ -1431,6 +1541,8 @@ xkb_state_notify_event_filter (
 #define group_change_mask (XkbGroupStateMask | XkbGroupBaseMask | XkbGroupLatchMask | XkbGroupLockMask)
 #define modifier_change_mask (XkbModifierStateMask | XkbModifierBaseMask | XkbModifierLatchMask | XkbModifierLockMask)
 
+  if (!drawing->xkb) return GDK_FILTER_CONTINUE;
+
   if (((XEvent *) gdkxev)->type == drawing->xkb_event_type)
     {
       XkbEvent *kev = (XkbEvent *) gdkxev;
@@ -1440,7 +1552,7 @@ xkb_state_notify_event_filter (
           if (((kev->state.changed & modifier_change_mask) &&
                drawing->track_modifiers))
             {
-              free_keys_and_doodads_and_colors (drawing);
+              free_cdik (drawing);
               if (drawing->track_modifiers)
                 keyboard_drawing_set_mods (drawing, kev->state.compat_state);
               drawing->keys =
@@ -1504,6 +1616,7 @@ static void
 destroy (
   KeyboardDrawing * drawing)
 {
+  free_pango_layout (drawing);
   gdk_window_remove_filter (NULL,
                             (GdkFilterFunc) xkb_state_notify_event_filter,
                             drawing);
@@ -1514,12 +1627,18 @@ destroy (
     }
 }
 
+static void
+style_changed (
+  KeyboardDrawing * drawing)
+{
+  pango_layout_context_changed (drawing->layout);
+}
+
 static void
 keyboard_drawing_init (
   KeyboardDrawing * drawing)
 {
   gint opcode = 0, error = 0, major = 1, minor = 0;
-  PangoContext *context;
   gint mask;
 
   drawing->display = GDK_DISPLAY_XDISPLAY (gdk_display_get_default ());
@@ -1537,9 +1656,8 @@ keyboard_drawing_init (
     drawing->screen_num = gdk_screen_get_number (gdk_screen_get_default ());
 
   drawing->pixmap = NULL;
-  context = pango_xft_get_context (drawing->display, drawing->screen_num);
-  drawing->layout = pango_layout_new (context);
-  g_object_unref (context);
+  alloc_pango_layout (drawing);
+
   drawing->font_desc =
     pango_font_description_copy (GTK_WIDGET (drawing)->style->font_desc);
   drawing->keyboard_items = NULL;
@@ -1559,16 +1677,16 @@ keyboard_drawing_init (
                                  XkbGBN_OtherNamesMask | XkbGBN_SymbolsMask |
                                  XkbGBN_IndicatorMapMask, XkbUseCoreKbd);
   if (drawing->xkb == NULL)
+  {
     g_critical ("XkbGetKeyboard failed to get keyboard from the server!");
+    return;
+  }
 
   XkbGetNames (drawing->display, XkbAllNamesMask, drawing->xkb);
 
   drawing->xkbOnDisplay = TRUE;
 
-  drawing->physical_indicators_size =
-    drawing->xkb->indicators->phys_indicators + 1;
-  drawing->physical_indicators =
-    g_new0 (KeyboardDrawingDoodad *, drawing->physical_indicators_size);
+  alloc_cdik (drawing);
 
   XkbSelectEventDetails (drawing->display, XkbUseCoreKbd,
                          XkbIndicatorStateNotify,
@@ -1587,7 +1705,6 @@ keyboard_drawing_init (
   mask = (XkbGroupNamesMask | XkbIndicatorNamesMask);
   XkbSelectEventDetails (drawing->display, XkbUseCoreKbd, XkbNamesNotify,
                          mask, mask);
-  drawing->keys = g_new0 (KeyboardDrawingKey, drawing->xkb->max_key_code + 1);
   init_keys_and_doodads (drawing);
   init_colors (drawing);
 
@@ -1614,6 +1731,8 @@ keyboard_drawing_init (
                     G_CALLBACK (size_allocate), drawing);
   g_signal_connect (G_OBJECT (drawing), "destroy", G_CALLBACK (destroy),
                     drawing);
+  g_signal_connect (G_OBJECT (drawing), "style-set", G_CALLBACK (style_changed),
+                    drawing);
 
   gdk_window_add_filter (NULL, (GdkFilterFunc) xkb_state_notify_event_filter,
                          drawing);
@@ -1668,16 +1787,16 @@ keyboard_drawing_get_type (void)
 
 void
 keyboard_drawing_set_mods (
-  KeyboardDrawing * kbdrawing,
+  KeyboardDrawing * drawing,
   guint mods)
 {
 #ifdef KBDRAW_DEBUG
   printf ("set_mods: %d\n", mods);
 #endif
-  if (mods != kbdrawing->mods)
+  if (mods != drawing->mods)
     {
-      kbdrawing->mods = mods;
-      gtk_widget_queue_draw (GTK_WIDGET (kbdrawing));
+      drawing->mods = mods;
+      gtk_widget_queue_draw (GTK_WIDGET (drawing));
     }
 }
 
@@ -1685,163 +1804,169 @@ keyboard_drawing_set_mods (
  * (which can then be saved to disk, etc) */
 GdkPixbuf *
 keyboard_drawing_get_pixbuf (
-  KeyboardDrawing * kbdrawing)
+  KeyboardDrawing * drawing)
 {
-  if (kbdrawing->pixmap == NULL)
-    draw_keyboard (kbdrawing);
+  if (drawing->pixmap == NULL)
+    draw_keyboard (drawing);
 
-  return gdk_pixbuf_get_from_drawable (NULL, kbdrawing->pixmap, NULL, 0, 0, 0,
-                                       0, xkb_to_pixmap_coord (kbdrawing,
-                                                               kbdrawing->
+  return gdk_pixbuf_get_from_drawable (NULL, drawing->pixmap, NULL, 0, 0, 0,
+                                       0, xkb_to_pixmap_coord (drawing,
+                                                               drawing->
                                                                xkb->geom->
                                                                width_mm),
-                                       xkb_to_pixmap_coord (kbdrawing,
-                                                            kbdrawing->xkb->
+                                       xkb_to_pixmap_coord (drawing,
+                                                            drawing->xkb->
                                                             geom->height_mm));
 }
 
 gboolean
 keyboard_drawing_set_keyboard (
-  KeyboardDrawing * kbdrawing,
+  KeyboardDrawing * drawing,
   XkbComponentNamesRec * names)
 {
-  free_keys_and_doodads_and_colors (kbdrawing);
-  XkbFreeKeyboard (kbdrawing->xkb, 0, TRUE);  /* free_all = TRUE */
-  kbdrawing->xkb = NULL;
+  free_cdik (drawing);
+  if (drawing->xkb)
+    XkbFreeKeyboard (drawing->xkb, 0, TRUE);  /* free_all = TRUE */
+  drawing->xkb = NULL;
 
   if (names)
     {
-      kbdrawing->xkb =
-        XkbGetKeyboardByName (kbdrawing->display, XkbUseCoreKbd, names, 0,
+      drawing->xkb =
+        XkbGetKeyboardByName (drawing->display, XkbUseCoreKbd, names, 0,
                               XkbGBN_GeometryMask | XkbGBN_KeyNamesMask |
                               XkbGBN_OtherNamesMask | XkbGBN_ClientSymbolsMask
                               | XkbGBN_IndicatorMapMask, FALSE);
-      kbdrawing->xkbOnDisplay = FALSE;
+      drawing->xkbOnDisplay = FALSE;
     }
   else
     {
-      kbdrawing->xkb = XkbGetKeyboard (kbdrawing->display,
-                                       XkbGBN_GeometryMask |
-                                       XkbGBN_KeyNamesMask |
-                                       XkbGBN_OtherNamesMask |
-                                       XkbGBN_SymbolsMask |
-                                       XkbGBN_IndicatorMapMask,
-                                       XkbUseCoreKbd);
-      XkbGetNames (kbdrawing->display, XkbAllNamesMask, kbdrawing->xkb);
-      kbdrawing->xkbOnDisplay = TRUE;
+      drawing->xkb = XkbGetKeyboard (drawing->display,
+                                     XkbGBN_GeometryMask |
+                                     XkbGBN_KeyNamesMask |
+                                     XkbGBN_OtherNamesMask |
+                                     XkbGBN_SymbolsMask |
+                                     XkbGBN_IndicatorMapMask,
+                                     XkbUseCoreKbd);
+      XkbGetNames (drawing->display, XkbAllNamesMask, drawing->xkb);
+      drawing->xkbOnDisplay = TRUE;
     }
 
-  if (kbdrawing->xkb == NULL)
+  if (drawing->xkb == NULL)
     return FALSE;
 
-  kbdrawing->keys =
-    g_new0 (KeyboardDrawingKey, kbdrawing->xkb->max_key_code + 1);
-  init_keys_and_doodads (kbdrawing);
-  init_colors (kbdrawing);
+  alloc_cdik (drawing);
 
-  size_allocate (GTK_WIDGET (kbdrawing),
-                 &(GTK_WIDGET (kbdrawing)->allocation), kbdrawing);
-  gtk_widget_queue_draw (GTK_WIDGET (kbdrawing));
+  init_keys_and_doodads (drawing);
+  init_colors (drawing);
+
+  size_allocate (GTK_WIDGET (drawing),
+                 &(GTK_WIDGET (drawing)->allocation), drawing);
+  gtk_widget_queue_draw (GTK_WIDGET (drawing));
 
   return TRUE;
 }
 
 G_CONST_RETURN gchar *
 keyboard_drawing_get_keycodes (
-  KeyboardDrawing * kbdrawing)
+  KeyboardDrawing * drawing)
 {
-  if (kbdrawing->xkb->names->keycodes <= 0)
+  if (!drawing->xkb ||
+      drawing->xkb->names->keycodes <= 0)
     return NULL;
   else
-    return XGetAtomName (kbdrawing->display, kbdrawing->xkb->names->keycodes);
+    return XGetAtomName (drawing->display, drawing->xkb->names->keycodes);
 }
 
 G_CONST_RETURN gchar *
 keyboard_drawing_get_geometry (
-  KeyboardDrawing * kbdrawing)
+  KeyboardDrawing * drawing)
 {
-  if (kbdrawing->xkb->names->geometry <= 0)
+  if (!drawing->xkb ||
+      drawing->xkb->names->geometry <= 0)
     return NULL;
   else
-    return XGetAtomName (kbdrawing->display, kbdrawing->xkb->names->geometry);
+    return XGetAtomName (drawing->display, drawing->xkb->names->geometry);
 }
 
 G_CONST_RETURN gchar *
 keyboard_drawing_get_symbols (
-  KeyboardDrawing * kbdrawing)
+  KeyboardDrawing * drawing)
 {
-  if (kbdrawing->xkb->names->symbols <= 0)
+  if (!drawing->xkb ||
+      drawing->xkb->names->symbols <= 0)
     return NULL;
   else
-    return XGetAtomName (kbdrawing->display, kbdrawing->xkb->names->symbols);
+    return XGetAtomName (drawing->display, drawing->xkb->names->symbols);
 }
 
 G_CONST_RETURN gchar *
 keyboard_drawing_get_types (
-  KeyboardDrawing * kbdrawing)
+  KeyboardDrawing * drawing)
 {
-  if (kbdrawing->xkb->names->types <= 0)
+  if (!drawing->xkb ||
+      drawing->xkb->names->types <= 0)
     return NULL;
   else
-    return XGetAtomName (kbdrawing->display, kbdrawing->xkb->names->types);
+    return XGetAtomName (drawing->display, drawing->xkb->names->types);
 }
 
 G_CONST_RETURN gchar *
 keyboard_drawing_get_compat (
-  KeyboardDrawing * kbdrawing)
+  KeyboardDrawing * drawing)
 {
-  if (kbdrawing->xkb->names->compat <= 0)
+  if (!drawing->xkb ||
+      drawing->xkb->names->compat <= 0)
     return NULL;
   else
-    return XGetAtomName (kbdrawing->display, kbdrawing->xkb->names->compat);
+    return XGetAtomName (drawing->display, drawing->xkb->names->compat);
 }
 
 void
 keyboard_drawing_set_track_modifiers (
-  KeyboardDrawing * kbdrawing,
+  KeyboardDrawing * drawing,
   gboolean enable)
 {
   if (enable)
     {
       XkbStateRec state;
-      kbdrawing->track_modifiers = 1;
+      drawing->track_modifiers = 1;
       memset (&state, 0, sizeof (state));
-      XkbGetState (kbdrawing->display, XkbUseCoreKbd, &state);
-      keyboard_drawing_set_mods (kbdrawing, state.compat_state);
+      XkbGetState (drawing->display, XkbUseCoreKbd, &state);
+      keyboard_drawing_set_mods (drawing, state.compat_state);
     }
   else
-    kbdrawing->track_modifiers = 0;
+    drawing->track_modifiers = 0;
 }
 
 void
 keyboard_drawing_set_track_config (
-  KeyboardDrawing * kbdrawing,
+  KeyboardDrawing * drawing,
   gboolean enable)
 {
   if (enable)
-    kbdrawing->track_config = 1;
+    drawing->track_config = 1;
   else
-    kbdrawing->track_config = 0;
+    drawing->track_config = 0;
 }
 
-void keyboard_drawing_set_groups_levels(KeyboardDrawing      *kbdrawing, 
+void keyboard_drawing_set_groups_levels(KeyboardDrawing      *drawing, 
   KeyboardDrawingGroupLevel * groupLevels[])
 {
 #ifdef KBDRAW_DEBUG
   printf ("set_group_levels [topLeft]: %d %d \n", 
-	groupLevels[KEYBOARD_DRAWING_POS_TOPLEFT].group, 
-	groupLevels[KEYBOARD_DRAWING_POS_TOPLEFT].level );
+	groupLevels[KEYBOARD_DRAWING_POS_TOPLEFT]->group, 
+	groupLevels[KEYBOARD_DRAWING_POS_TOPLEFT]->level );
   printf ("set_group_levels [topRight]: %d %d \n", 
-	groupLevels[KEYBOARD_DRAWING_POS_TOPRIGHT].group, 
-	groupLevels[KEYBOARD_DRAWING_POS_TOPRIGHT].level );
+	groupLevels[KEYBOARD_DRAWING_POS_TOPRIGHT]->group, 
+	groupLevels[KEYBOARD_DRAWING_POS_TOPRIGHT]->level );
   printf ("set_group_levels [bottomLeft]: %d %d \n", 
-	groupLevels[KEYBOARD_DRAWING_POS_BOTTOMLEFT].group, 
-	groupLevels[KEYBOARD_DRAWING_POS_BOTTOMLEFT].level );
+	groupLevels[KEYBOARD_DRAWING_POS_BOTTOMLEFT]->group, 
+	groupLevels[KEYBOARD_DRAWING_POS_BOTTOMLEFT]->level );
   printf ("set_group_levels [bottomRight]: %d %d \n", 
-	groupLevels[KEYBOARD_DRAWING_POS_BOTTOMRIGHT].group, 
-	groupLevels[KEYBOARD_DRAWING_POS_BOTTOMRIGHT].level );
+	groupLevels[KEYBOARD_DRAWING_POS_BOTTOMRIGHT]->group, 
+	groupLevels[KEYBOARD_DRAWING_POS_BOTTOMRIGHT]->level );
 #endif
-  kbdrawing->groupLevels = groupLevels;
+  drawing->groupLevels = groupLevels;
 
-  gtk_widget_queue_draw (GTK_WIDGET (kbdrawing));
+  gtk_widget_queue_draw (GTK_WIDGET (drawing));
 }
